import json
import datetime
from bson import objectid


def toJSON(data):
    return json.dumps(data, ensure_ascii=False, cls=genericJsonEncoder)


class genericJsonEncoder(json.JSONEncoder):
    def default(self, data):
        if isinstance(data, objectid.ObjectId):
            return str(data)
        elif isinstance(data, datetime.datetime):
            return data.strftime("%Y-%m-%dT%H:%MZ")
        else:
            return json.JSONEncoder.default(self, data)



from flask import Response, request
from apicore import api, Http501Exception, Http403Exception


@api.route('/oas/test/', methods=['POST'])
@api.validate
def aosTest():
    """
    common_responses:
        400:
         description: Invalid request
        401:
          description: Authentification required
        406:
          description: Nothing to send maching Access-* headers
        500:
          description: Server internal error
    ---
      tags:
        - transation
      summary: Perfom a transaction by using a code
      requestBody:
        content:
          application/json:
            schema:
              type: object
              required:
                - amount
                - code
              properties:
                amount:
                  title: Montant en Francs CFA à dépenser chez le commerçant
                  type: number
                  example: 35000.00
                code:
                  title: Code du destinataire utilisé pour la transaction
                  type: string
                  example: 35-768-99
        required: true
      responses:
        201:
          description: Created
          content:
            application/json:
              schema:
                type: object
                properties:
                  uuid:
                    title: Unique identifier of the created transaction
                    type: string
                    format: uuid
                    example: 59f226c110cd249a8d4a7fd7
        403:
          description: Max amount exceeded
        404:
            description: Invalid code
    """
    return "OK!!"



################################################################################
# MIT License
#
# Copyright (c) 2016 Meezio SAS <dev@meez.io>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
################################################################################

from flask import request


class Lang:
    @staticmethod
    def best_match(available_languages, default=None):
        """ Determine best language from list of available languages and from Accept-Language Header.

        :param list available_languages: List of available languages to match with.
        :param str default: Language returned when no match found.
        :return str: The best matching language or None if no matching.
        """
        for lang in request.accept_languages.values():
            if lang in available_languages:
                return lang
            elif len(lang) > 2 and lang[:2] in available_languages:
                return lang[:2]

        return default



import jsonschema


def validate(data, schema):
    try:
        jsonschema.validate(data, schema)
    except jsonschema.exceptions.ValidationError:
        return False

    # TODO if $ref or
    if "type" in schema and schema["type"] == "object":
        allowed = list(schema["properties"].keys())
        for key in data.keys():
            if key not in allowed:
                return False

    return True



import yaml
import re
from flask import request
from .logger import Logger
from .exceptions import Http406Exception, Http400Exception
from .oasschema import validate


class OpenAPI:
    def __init__(self, appname, endpoint="/openapi.json", openApiFile="conf/openapi.yaml"):
        self.endpoint = endpoint
        self._pattern = re.compile(r"<([^<]*:)?([^<]*)>")
        self._replace = r"{\2}"
        self._doc = dict()

        self.spec = dict()
        self.spec["info"] = dict()
        self.spec["paths"] = dict()

        try:
            with open(openApiFile) as f:
                data = yaml.load(f)
                if data:
                    self.spec.update(data)
                    if "title" not in self.spec["info"]:
                        self.spec["info"]["title"] = "{} API Specifications".format(appname)
                    if "version" not in self.spec["info"]:
                        self.spec["info"]["version"] = "1.0.0"
                Logger.info("'{}' loaded".format(openApiFile))
        except FileNotFoundError:
            pass

        self.spec["openapi"] = "3.0.0"

    def addEndpoint(self, rule, methods, docstring, funcName):
        if docstring:
            clean = re.sub(self._pattern, self._replace, rule)
            self.spec["paths"][clean] = dict()
            common_responses = dict()
            docs = docstring.split("---")

            try:
                once = True
                idx = 0
                for doc in docs:
                    yl = yaml.load(doc)
                    if "responses" in yl:
                        yl["responses"].update(common_responses)
                        self.spec["paths"][clean][methods[idx].lower()] = yl
                        if idx < len(methods) - 1:
                            idx += 1
                    elif once:
                        common_responses = yl.pop('common_responses', dict())
                        self.spec["paths"][clean] = yl
                    once = False
                self._doc[funcName] = self.spec["paths"][clean]
            except:
                Logger.error('"Invalide OpenAPI Spec for endpoint "{}" {}'.format(rule, methods))

    def check(self, funcName, method, *args, **kwargs):
        schemas = self.spec.get("components", {})

        try:
            doc = self._doc[funcName][method.lower()]
        except:
            Logger.error("OpenAPI Specifications not found for {} on route def ''{}()''".format(method, funcName))
            return

        body = doc.get("requestBody", None)
        if body:
            dataRequired = body.get("required", False)

            keys = body["content"].keys()
            if len(keys) > 1:
                Logger.error("OpenAPI Specifications marlformed body content for {} on route def '{}()'".format(method, funcName))
                return
            contentType = list(keys)[0]
            schema = (body["content"][contentType]).get("schema", None)

            if not schema:
                Logger.error("OpenAPI Specifications body content schema missing for {} on route def '{}()'".format(method, funcName))
                return

            # NOTE if contentType != Accept in request.headers, return Http406Exception
            schema["components"] = schemas
            if contentType == "application/json":
                data = request.get_json()
                if (not data and dataRequired) or not validate(data, schema):
                    return Http400Exception

        # print(doc)
        # TODO validate query (/?name=toto) 400
        # TODO validate args (/<args>/) 400 & *args, **kwargs
        # if "parameters" in self._doc[funcName][method.lower()]:
        #    pass

    def fake(self, funcName, method):
        try:
            # TODO parcourir la doc pour trouver une request valide (reponse 200 ou < 300)
            return str(self._doc[funcName][method.lower()]["responses"])
        except:
            # TODO raise error 500 : OpenAPI Specifications not found
            return "Fake data"
